package edu.sysu.pmglab.downloader;

import edu.sysu.pmglab.check.Assert;
import edu.sysu.pmglab.commandParser.CommandGroup;
import edu.sysu.pmglab.commandParser.CommandOption;
import edu.sysu.pmglab.commandParser.CommandOptions;
import edu.sysu.pmglab.commandParser.CommandParser;
import edu.sysu.pmglab.commandParser.types.FILE;
import edu.sysu.pmglab.commandParser.types.INTEGER;
import edu.sysu.pmglab.commandParser.types.IType;
import edu.sysu.pmglab.commandParser.types.STRING;
import edu.sysu.pmglab.commandParser.usage.DefaultStyleUsage;
import edu.sysu.pmglab.container.File;
import edu.sysu.pmglab.easytools.StringUtils;
import edu.sysu.pmglab.threadPool.ThreadPool;
import edu.sysu.pmglab.unifyIO.FileStream;
import edu.sysu.pmglab.unifyIO.InputStreamWrapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.*;
import java.util.ArrayList;
import java.util.Base64;
import java.util.List;
import java.util.concurrent.Future;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

import static edu.sysu.pmglab.commandParser.CommandItem.HELP;
import static edu.sysu.pmglab.commandParser.CommandItem.HIDDEN;

/**
 * @author suranyi
 */

public class HttpDownloader {
    private static final Logger logger = LoggerFactory.getLogger("HttpDownloader");

    final String url;
    final File outputFile;
    final int nThreads;
    final File tempDir;
    final boolean printLog;
    final boolean printProcess;
    final Proxy proxy;
    final int timeOut;
    final boolean clean;
    final boolean resume;

    final AtomicLong downloadSize = new AtomicLong(0);
    final AtomicInteger finishThread = new AtomicInteger(0);
    final long httpContentLength;

    public static class Proxy {
        final String host;
        final String port;

        public Proxy(String host, String port) {
            Assert.NotNull(host);
            Assert.NotNull(port);
            this.host = host;
            this.port = port;
        }

        public Proxy(String host_port) {
            Assert.NotNull(host_port);
            String[] split = host_port.split(":");
            this.host = split[0];
            this.port = split[1];
        }
    }

    public static class Range {
        long start;
        long end;

        public Range(long start, long end) {
            Assert.valueRange(start, 0, Long.MAX_VALUE);

            // end 为 -1 代表不设定末端范围
            Assert.valueRange(end, -1, Long.MAX_VALUE);
            this.start = start;
            this.end = end;
        }

        public Range(long start) {
            this(start, -1);
        }
    }

    /**
     * 构建器
     */
    public static class Builder {
        final String url;
        File outputFile = null;
        int nThreads = 5;
        File tempDir = null;
        boolean printLog = false;
        boolean printProcess = false;
        Proxy proxy = null;
        int timeOut = 10;
        boolean clean = false;

        private Builder(String url) throws UnsupportedEncodingException {
            if (url.startsWith("thunder://")) {
                // 迅雷链接转换
                url = url.substring(10);
                // base 64 转换
                url = new String(Base64.getDecoder().decode(url.getBytes()), "UTF-8");
                // url 解码
                url = URLDecoder.decode(url, "UTF-8");

                // 去头去尾
                if (url.startsWith("AA")) {
                    url = url.substring(2);
                }
                if (url.endsWith("ZZ")) {
                    url = url.substring(0, url.length() - 2);
                }
            }

            this.url = url;
        }

        /**
         * 设置并行下载
         *
         * @param nThreads 并行线程数
         */
        public Builder setThreads(int nThreads) {
            this.nThreads = Math.max(nThreads, 1);
            return this;
        }

        /**
         * 设置输出文件路径
         *
         * @param outputFile 输出文件
         */
        public Builder setOutputFile(File outputFile) {
            if (outputFile != null) {
                if (outputFile.isExists() && outputFile.isDirectory()) {
                    this.outputFile = outputFile.getSubFile(new File(this.url).getFileName());
                } else {
                    this.outputFile = outputFile;
                }
            } else {
                this.outputFile = null;
            }
            return this;
        }

        /**
         * 设置输出文件路径
         *
         * @param outputFileName 输出文件名
         */
        public Builder setOutputFile(String outputFileName) {
            if (outputFileName == null) {
                this.outputFile = null;
                return this;
            } else {
                return setOutputFile(new File(outputFileName));
            }
        }

        /**
         * 设置临时文件路径
         *
         * @param tempDir 临时文件路径
         */
        public Builder setTempDir(File tempDir) {
            this.tempDir = tempDir;
            return this;
        }

        /**
         * 设置临时文件路径
         *
         * @param tempDirName 临时文件路径
         */
        public Builder setTempDir(String tempDirName) {
            if (tempDirName == null) {
                this.tempDir = null;
                return this;
            } else {
                return setTempDir(new File(tempDirName));
            }
        }

        /**
         * 是否打印程序日志
         * 当 Downloader 作为单独的软件包使用时, 建议打开日志记录
         *
         * @param printLog 是否使用日志
         */
        public Builder setPrintLog(boolean printLog) {
            this.printLog = printLog;
            this.printProcess = printLog;
            return this;
        }

        /**
         * 是否打印下载进度
         *
         * @param printProcess 是否使用日志
         */
        public Builder setPrintProcess(boolean printProcess) {
            this.printProcess = printProcess;
            return this;
        }

        /**
         * 清除缓存数据
         *
         * @param clean 是否清除缓存数据
         */
        public Builder clean(boolean clean) {
            this.clean = clean;
            return this;
        }

        /**
         * 设置代理
         *
         * @param host 代理地址, 通常为 127.0.0.1
         * @param port 代理端口, 取决于本机代理 socket 端口 (mac 上 clashX 通常为 7890)
         */
        public Builder setProxy(String host, String port) {
            if (host == null) {
                host = "127.0.0.1";
            }

            if (port == null) {
                this.proxy = null;
            } else {
                this.proxy = new Proxy(host, port);
            }
            return this;
        }

        /**
         * 设置代理
         *
         * @param host_port 代理地址, 通常为 127.0.0.1:7890
         */
        public Builder setProxy(String host_port) {
            if (host_port == null) {
                this.proxy = null;
            } else {
                this.proxy = new Proxy(host_port);
            }
            return this;
        }

        /**
         * 设置代理
         *
         * @param proxy 代理对象
         */
        public Builder setProxy(Proxy proxy) {
            this.proxy = proxy;
            return this;
        }

        /**
         * 设置最大等待时间
         *
         * @param timeOut 等待时间
         */
        public Builder setTimeOut(int timeOut) {
            this.timeOut = Math.max(timeOut, 1);
            return this;
        }

        /**
         * 提交下载任务
         */
        public boolean download() throws IOException {
            // 进行 url 跳转
            return new HttpDownloader(this).download();
        }
    }

    public static void main(String[] args) {
        try {
            if (args.length == 0) {
                System.out.println(HttpDownloaderParser.getParser());
                return;
            }

            HttpDownloaderParser options = HttpDownloaderParser.parse(args);
            if (options.help.isPassedIn) {
                System.out.println(HttpDownloaderParser.getParser());
                return;
            }

            do {
                try {
                    HttpDownloader.instance(args[0])
                            .setOutputFile(options.output.value)
                            .setThreads(options.threads.value)
                            .setPrintLog(true)
                            .setTempDir(options.tempDir.value)
                            .setProxy(options.proxy.value)
                            .setTimeOut(options.timeout.value)
                            .clean(options.overwrite.isPassedIn)
                            .download();
                } catch (IOException e) {
                    logger.error("{}", e.getMessage());

                    if (!options.noAutoRetry.isPassedIn) {
                        logger.error("Download failed, resume download in 3 seconds.");

                        Thread.sleep(3000);
                        continue;
                    }
                }

                break;
            } while (true);
        } catch (Exception | Error e) {
            logger.error("{}", e.getMessage());
        }
    }

    public static Builder instance(String url) throws UnsupportedEncodingException {
        return new Builder(url);
    }

    public static boolean hasNewVersion(String url, File to) throws IOException {
        HttpURLConnection connection = HttpDownloader.getHttpURLConnection(url, null, null, 5);
        long size = connection.getContentLengthLong();
        connection.disconnect();
        long localSize = 0;

        if (to.isDirectory()) {
            to = to.getSubFile(new File(url).getFileName());
        }

        if (!to.isDirectory()) {
            localSize = to.size();
        }

        return size <= 0 || size != localSize;
    }

    public static void updateResource(String url, File to, String softwareName) throws IOException {
        if (softwareName == null) {
            softwareName = "software";
        }

        if (!hasNewVersion(url, to)) {
            logger.info("The current version of {} is the latest version!", softwareName);
        } else {
            logger.info("New version of the {} is available, start downloading.", softwareName);
            HttpDownloader.instance(url)
                    .setOutputFile(to)
                    .setThreads(1)
                    .setPrintLog(false)
                    .setPrintProcess(true)
                    .download();
            logger.info("Download completed.");
        }
    }

    private HttpDownloader(Builder builder) throws IOException {
        this.timeOut = builder.timeOut;
        this.url = redirectURL(builder.url, builder.proxy, this.timeOut);
        this.outputFile = builder.outputFile == null ? new File("./" + new File(this.url).getFileName()) : builder.outputFile;
        this.tempDir = builder.tempDir == null ? this.outputFile.addExtension(".~$temp") : builder.tempDir;
        this.printLog = builder.printLog;
        this.proxy = builder.proxy;
        this.printProcess = builder.printProcess;

        HttpURLConnection connection = getHttpURLConnection(this.url, builder.proxy, new Range(0), this.timeOut);
        this.httpContentLength = connection.getContentLengthLong();

        try {
            if (connection.getResponseCode() == 404) {
                throw new IOException("Resource not found.");
            }

            if (!String.valueOf(connection.getResponseCode()).startsWith("2")) {
                throw new IOException("Network connection error, response code: " + connection.getResponseCode());
            }
        } catch (UnknownHostException e) {
            throw new IOException("Network connection interrupted.");
        }

        if (connection.getResponseCode() != 206) {
            // 不支持随机读取
            this.nThreads = 1;
            this.resume = false;

            // 必须清除缓存数据
            this.clean = true;
        } else {
            if (this.httpContentLength == -1) {
                // 虽然支持跳转, 但是没有明确的长度
                this.nThreads = 1;
            } else {
                this.nThreads = builder.nThreads;
            }

            this.resume = true;
            this.clean = builder.clean;
        }
        connection.disconnect();
    }

    public boolean download() throws IOException {
        if (this.clean) {
            if (this.tempDir.isExists()) {
                for (int i = 0; i < this.nThreads; i++) {
                    File tempFile = tempDir.getSubFile("part" + i + ".~$temp");
                    if (tempFile.isExists()) {
                        tempFile.delete();
                    }
                }
            }

            if (this.outputFile.isExists()) {
                this.outputFile.delete();
            }
        }

        // 开始下载任务,  预留一个日志线程
        ThreadPool pool = new ThreadPool(this.nThreads + 1);

        // 记录开始时间
        long startTime = System.currentTimeMillis();

        // 本地文件大小
        long localFileSize = this.outputFile.size();

        // 获取网络文件具体大小
        if (this.printLog) {
            logger.info("Start download file {} (from {})", outputFile, this.url);
        }

        if (localFileSize > 0 && localFileSize == this.httpContentLength) {
            if (this.printLog) {
                String format = String.format("> Downloaded: %s / %s (%.1f %%); Speed: %s; Time Left: %s s", formatSize(localFileSize, 1), formatSize(httpContentLength, 1), 100f, formatSpeed((double) this.httpContentLength / ((System.currentTimeMillis() - startTime) / 1000d), 1), 0);
                System.out.println(format);
            }
        } else {
            if (localFileSize > 0) {
                // 此时删除本地文件, 重新下载
                this.outputFile.delete();
                localFileSize = 0;
            }

            if (!resume && this.printLog) {
                logger.info("Current url doesn't support resume download.");
            }

            // 添加已经下载部分的文件大小
            downloadSize.addAndGet(localFileSize);

            // 任务列表
            List<Future<Boolean>> futureList = new ArrayList<>();

            // 开始并行下载
            this.tempDir.mkdir();
            parallelDownload(localFileSize, pool, futureList);
            listenDownloadProcess(pool, futureList);

            // 阻塞等待
            pool.close();

            // 如果线程池发生了错误, 则抛出异常栈
            if (pool.isError()) {
                throw new IOException(pool.getCause().get(0));
            }

            // 文件合并及清理
            flush();
        }

        if (this.printLog) {
            logger.info("Download completed. Total time: {} s; File size: {}", (System.currentTimeMillis() - startTime) / 1000d, this.outputFile.formatSize(3));
        }

        return true;
    }

    private void flush() throws IOException {
        if (this.printLog) {
            logger.info("Merge temporary files to file {}", this.outputFile);
        }

        if (nThreads == 1) {
            tempDir.getSubFile("part0.~$temp").rename(outputFile);

        } else {
            FileStream writer = this.outputFile.open(FileStream.CHANNEL_WRITER);

            for (int i = 0; i < nThreads; i++) {
                FileStream tempFile = tempDir.getSubFile("part" + i + ".~$temp").open(FileStream.CHANNEL_READER);
                tempFile.writeTo(0, tempFile.size(), writer.getChannel());
                tempFile.close();
            }
            writer.close();

            // 最终删除所有子文件
            for (int i = 0; i < this.nThreads; i++) {
                File tempFile = tempDir.getSubFile("part" + i + ".~$temp");
                if (tempFile.isExists()) {
                    tempFile.delete();
                }
            }

        }
        if (tempDir.size() == 0) {
            this.tempDir.delete();
        }
    }

    /**
     * 并行下载, 按照原大小近似切分为多个等长的部分
     */
    private void parallelDownload(long localFileSize, ThreadPool pool, List<Future<Boolean>> futureList) {
        if (!this.resume) {
            // 切分文件
            futureList.add(pool.submit(() -> download(null, 0)));
        } else {
            // 切分文件
            if (this.nThreads == 1) {
                futureList.add(pool.submit(() -> download(new Range(localFileSize, this.httpContentLength), 0)));
            } else {
                long needToDownload = this.httpContentLength - localFileSize;
                long size = (needToDownload) / this.nThreads;
                long lastSize = needToDownload - (needToDownload / this.nThreads * (this.nThreads - 1));

                for (int i = 0; i < this.nThreads; i++) {
                    long start = localFileSize + i * size;
                    long downloadWindow = (i == this.nThreads - 1) ? lastSize : size;
                    long end = start + downloadWindow;
                    if (start != 0) {
                        start++;
                    }

                    Range range = new Range(start, end);
                    int partIndex = i;
                    futureList.add(pool.submit(() -> download(range, partIndex)));
                }
            }
        }
    }

    /**
     * 监听下载过程
     */
    private void listenDownloadProcess(ThreadPool pool, List<Future<Boolean>> futureList) {
        if (!this.printProcess) {
            futureList.add(pool.submit(() -> {
                long lastSize = 0;
                int noChangeTime = 0;

                while (finishThread.get() != this.nThreads) {
                    long downloadSize = HttpDownloader.this.downloadSize.get();
                    long changeSize = downloadSize - lastSize;
                    lastSize = downloadSize;

                    if (changeSize == 0) {
                        noChangeTime++;
                    } else {
                        noChangeTime = 0;
                    }

                    Thread.sleep(1000);

                    if (noChangeTime >= timeOut) {
                        throw new TimeoutException("Current session is time out, please download it again.");
                    }
                }
                return true;
            }));
        } else {
            futureList.add(pool.submit(() -> {
                double lastSize = 0;
                int noChangeTime = 0;

                // 文件总大小
                String totalSize;
                if (this.httpContentLength == -1) {
                    totalSize = "-";
                } else {
                    totalSize = formatSize(this.httpContentLength, 1);
                }

                while (finishThread.get() != this.nThreads) {
                    long downloadSize = HttpDownloader.this.downloadSize.get();

                    // 每秒速度 (单位: B / s)
                    int speed = (int) (downloadSize - lastSize);

                    // 剩余时间
                    String surplusTime;
                    if (this.httpContentLength == -1) {
                        surplusTime = "-";
                    } else {
                        double surplusSize = this.httpContentLength - downloadSize;
                        surplusTime = speed == 0 ? "-" : String.format("%.1f", surplusSize / speed);
                    }

                    // 已下百分比
                    String downloadRare = this.httpContentLength == -1 ? "-" : String.format("%.1f", ((double) downloadSize / this.httpContentLength) * 100d);
                    String speedFormat = formatSpeed(speed, 1);
                    String format = String.format("Downloaded: %s / %s (%s %%); Speed: %s; Time Left: %s s", formatSize(downloadSize, 1), totalSize, downloadRare, speedFormat, surplusTime);
                    System.out.print("\r");
                    System.out.print("/ " + format);
                    Thread.sleep(500);

                    System.out.print("\r");
                    System.out.print("| " + format);
                    Thread.sleep(500);

                    lastSize = downloadSize;

                    if (speed == 0) {
                        noChangeTime++;
                    } else {
                        noChangeTime = 0;
                    }

                    if (noChangeTime >= timeOut) {
                        System.out.println();
                        throw new TimeoutException("Current session is time out, please download it again.");
                    }

                    System.out.print("\r");
                    System.out.print(StringUtils.copyN(" ", format.length() + 2));
                }

                System.out.print("\r");
                String format = String.format("> Downloaded: %s / %s (%.1f %%); Speed: %s; Time Left: %s s", formatSize(downloadSize.get(), 1), formatSize(downloadSize.get(), 1), 100f, 0, 0);
                System.out.println(format);
                return true;
            }));
        }
    }

    private String formatSize(long size, int decimal) {
        String format = "%." + decimal + "f";
        if (size < 1024L) {
            return String.format("%d B", size);
        } else if (size < 1048576L) {
            return String.format(format + " KB", (float) size / 1024d);
        } else if (size < 1073741824L) {
            return String.format(format + " MB", (float) size / 1024d / 1024d);
        } else {
            return size < 1099511627776L ? String.format(format + " GB", (float) size / 1024d / 1024d / 1024d) : String.format(format + " TB", (float) size / 1024d / 1024d / 1024d / 1024d);
        }
    }

    private String formatSpeed(double speed, int decimal) {
        String format = "%." + decimal + "f";
        if (speed < 1024L) {
            return String.format(format + " B/s", speed);
        } else if (speed < 1048576L) {
            return String.format(format + " KB/s", speed / 1024d);
        } else if (speed < 1073741824L) {
            return String.format(format + " MB/s", speed / 1024d / 1024d);
        } else {
            return speed < 1099511627776L ? String.format(format + " GB/s", speed / 1024d / 1024d / 1024d) : String.format(format + " TB/s", speed / 1024d / 1024d / 1024d / 1024d);
        }
    }

    /**
     * 下载指定文件指定范围的数据
     *
     * @param range 指针范围
     * @param part  第几部分
     * @return true 时该线程下载完成
     */
    private boolean download(Range range, int part) throws IOException {
        // 文件名
        File tempFile = tempDir.getSubFile("part" + part + ".~$temp");

        // 本地切分子部分文件大小
        long localFileContentLength = tempFile.size();
        downloadSize.addAndGet(localFileContentLength);

        // 该线程已完成, 直接结束线程任务
        HttpURLConnection httpUrlConnection;
        if (range != null) {
            // 支持断点续传
            if (localFileContentLength >= range.end - range.start) {
                finishThread.addAndGet(1);
                return true;
            }

            if (this.printLog) {
                logger.info("Download file from {} to {} (file pointer range: {} - {})", this.url, tempFile, range.start, range.end);
            }

            // 创建 http 连接
            range.start += localFileContentLength;
            httpUrlConnection = getHttpURLConnection(this.url, this.proxy, range, this.timeOut);
        } else {
            // range == null, 不支持断点续传, 此时重新下载该文件
            if (this.printLog) {
                logger.info("Download file from {} to {}", this.url, tempFile);
            }

            // 创建 http 连接
            httpUrlConnection = getHttpURLConnection(this.url, this.proxy, null, this.timeOut);
        }

        // 获得输入流
        InputStream httpConnection = httpUrlConnection.getInputStream();
        InputStreamWrapper wrapper = new InputStreamWrapper(httpConnection);
        FileStream writer = tempFile.open(FileStream.DEFAULT_APPEND);

        try {
            byte[] buffer = new byte[8192 * 10];
            int len;

            while ((len = wrapper.read(buffer)) != -1) {
                writer.write(buffer, 0, len);
                downloadSize.addAndGet(len);
            }

            writer.close();
        } finally {
            httpUrlConnection.disconnect();
            finishThread.addAndGet(1);
        }
        return true;
    }

    /**
     * 获取 HTTP 链接
     */
    public static HttpURLConnection getHttpURLConnection(String url, Proxy proxy, Range range, int timeOut) throws IOException {
        URL httpUrl = new URL(url);

        HttpURLConnection httpConnection;
        if (proxy != null) {
            httpConnection = (HttpURLConnection) httpUrl.openConnection(new java.net.Proxy(java.net.Proxy.Type.HTTP, new InetSocketAddress(proxy.host, Integer.parseInt(proxy.port))));
        } else {
            httpConnection = (HttpURLConnection) httpUrl.openConnection();
        }

        httpConnection.setRequestProperty("User-Agent", "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36");

        if (range != null) {
            if (range.end == -1) {
                httpConnection.setRequestProperty("RANGE", "bytes=" + range.start + "-");
            } else {
                httpConnection.setRequestProperty("RANGE", "bytes=" + range.start + "-" + range.end);
            }
        }

        httpConnection.setConnectTimeout(1000 * timeOut);
        httpConnection.setReadTimeout(1000 * timeOut);
        return httpConnection;
    }

    /**
     * url 重定向
     *
     * @param url   URL 地址
     * @param proxy 代理
     */
    public static String redirectURL(String url, Proxy proxy, int timeOut) throws IOException {
        // 重定向跳转
        while (true) {
            HttpURLConnection httpConnection = getHttpURLConnection(url, proxy, null, timeOut);

            // 链接跳转
            String location = httpConnection.getHeaderField("Location");

            if (location == null) {
                httpConnection.disconnect();
                break;
            }

            url = location;
        }

        return url;
    }

    private static class HttpDownloaderParser {
        /**
         * build by: CommandParser-1.1
         * time: 2022-05-31 21:35:41
         */
        private static final CommandParser PARSER = new CommandParser(false);

        private final CommandOptions options;
        public final CommandOption<?> help;
        public final CommandOption<File> output;
        public final CommandOption<File> tempDir;
        public final CommandOption<Integer> threads;
        public final CommandOption<?> overwrite;
        public final CommandOption<String> proxy;
        public final CommandOption<Integer> timeout;
        public final CommandOption<?> noAutoRetry;

        HttpDownloaderParser(String... args) {
            this.options = PARSER.parse(args);
            this.help = new CommandOption<>("--help", this.options);
            this.output = new CommandOption<>("--output", this.options);
            this.tempDir = new CommandOption<>("--temp-dir", this.options);
            this.threads = new CommandOption<>("--threads", this.options);
            this.overwrite = new CommandOption<>("--overwrite", this.options);
            this.proxy = new CommandOption<>("--proxy", this.options);
            this.timeout = new CommandOption<>("--timeout", this.options);
            this.noAutoRetry = new CommandOption<>("--no-auto-retry", this.options);
        }

        public static HttpDownloaderParser parse(String... args) {
            return new HttpDownloaderParser(args);
        }

        public static HttpDownloaderParser parse(File argsFile) throws IOException {
            return new HttpDownloaderParser(CommandParser.readFromFile(argsFile));
        }

        /**
         * Get CommandParser
         */
        public static CommandParser getParser() {
            return PARSER;
        }

        /**
         * Get the usage of CommandParser
         */
        public static String usage() {
            return PARSER.toString();
        }

        /**
         * Get CommandOptions
         */
        public CommandOptions getOptions() {
            return this.options;
        }

        static {
            PARSER.setProgramName("<URL>");
            PARSER.offset(1);
            PARSER.debug(false);
            PARSER.usingAt(true);
            PARSER.setMaxMatchedNum(-1);
            PARSER.setAutoHelp(false);
            PARSER.setUsageStyle(new DefaultStyleUsage("Usage: ", " [options]", "About: Download file from an URL. The program supports breakpoints, parallel\n       downloads, proxy IP settings and is suitable for downloading a wide\n       range of http and thunder type urls.", 2, 4, 80, true, "*", "^"));

            CommandGroup group001 = PARSER.addCommandGroup("Options");
            group001.register(IType.NONE, "--help", "-help", "-h")
                    .addOptions(HELP, HIDDEN);
            group001.register(FILE.VALUE, "--output", "-o")
                    .setFormat("--output <file>")
                    .setDescription("Set the output file name, the default is inferred from the URL address.");
            group001.register(FILE.VALUE, "--temp-dir")
                    .setFormat("--temp-dir <file>")
                    .setDescription("Set the cache data path.");
            group001.register(INTEGER.VALUE, "--threads", "-t")
                    .defaultTo("5")
                    .validateWith(INTEGER.validateWith(1))
                    .setFormat("--threads <int>")
                    .setDescription("Number of threads for parallel download.");
            group001.register(IType.NONE, "--overwrite")
                    .setDescription("Overwrite existing files with --overwrite.");
            group001.register(STRING.VALUE, "--proxy")
                    .setFormat("--proxy host_ip:port")
                    .setDescription("Set the proxy to access the target resource.");
            group001.register(INTEGER.VALUE, "--timeout")
                    .defaultTo("10")
                    .validateWith(INTEGER.validateWith(1))
                    .setDescription("Disconnect if the server does not respond for more than --timeout second(s).");
            group001.register(IType.NONE, "--no-auto-retry")
                    .setDescription("Disable automatic reconnection on download failure.");
        }
    }
}
